// Copyright (C) 2001 Fabio Lopiano <flopiano@bigfoot.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


//
// Buffer class with reference counting, copy on write 
// and non-conservative resizing
//

 
#ifndef BUFFER_H
#define BUFFER_H

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>   // size_t

class Buffer
{
    class Rep
    {
         char * theBuffer;
         size_t theSize;
         int    theRefCnt;
         
         Rep(const Rep &rep);             // disabled;
         Rep & operator=(const Rep &rep);  // disabled;

         inline ~Rep() 
         {              
             assert(selfish());
             delete [] theBuffer; 
         }     

         friend class to_avoid_warning; // cause destructor is private

         inline void self_resize(int s) // only if selfish
         {
             assert(selfish());
             delete [] theBuffer;         

             if(theSize = s) // assignment
                 theBuffer = new char[ theSize ];
             else
                 theBuffer = 0;
         }
         
    public:

         inline Rep() : theBuffer(0),theSize(0), theRefCnt(1) {}
         inline Rep(size_t s) : 
             theBuffer(new char[s]),
	     theSize(s), 
             theRefCnt(1) {}

	     // buffer modification allowed
         inline bool selfish() { return theRefCnt == 1; }  

         inline size_t size() const { return theSize; }

	     // a read-only copy is returned 
         inline Rep * copy() 
         {
             theRefCnt++;
             return this;
         }

          // a changeable copy is returned 
         inline Rep * clone()
         {
             if(selfish()) { return this; }

             theRefCnt--;
        
             Rep *c = new Rep(theSize);
             memcpy(c->theBuffer,theBuffer,theSize);
             return c;
         }
         
         // one client less is referencing us
         inline void dismiss() 
         {  
             if(selfish()) 
                 delete this; 
             else 
                 theRefCnt --;
         }

         inline Rep * resize(int s) 
         {                      
            if(s == size()) return this; // nothing to do

            if(selfish()) { self_resize(s); return this; } // just resize ourself
                
            theRefCnt--;          // one client less is referencing us
            if(s)
            {
                return new Rep(s); // return a new rep with the new size
            }
            return new Rep;       // the new size is 0            
         }

         inline char * buffer() 
         { 
             assert(selfish()) ; 
             return theBuffer; 
         }
         
         inline const char *const_buffer() const 
         { 
             return theBuffer; 
         }
    };
                                
    Rep * theRep;

protected:
    // Copy On Write - call this function before any possible modification
    inline void cow() { theRep = theRep->clone(); }
    
    void copy(const Buffer &aBuffer)
    {
        if(theRep == aBuffer.theRep) return;

        theRep->dismiss();
        theRep = aBuffer.theRep->copy();
    }

public:   
    
    Buffer ();
    explicit Buffer (size_t aSize);
    Buffer (const char *aCharBuffer, size_t aSize);
	Buffer (const unsigned char *aCharBuffer, size_t aSize);
    Buffer (const Buffer &aBuffer);

    virtual ~Buffer() { theRep->dismiss(); }

    inline operator const char *() const { return theRep->const_buffer();   } 
    inline char *   buffer()             { cow(); return  theRep->buffer(); } 

    inline size_t size() const { return theRep->size(); }

    inline void resize (unsigned int aSize) { theRep = theRep->resize(aSize); }
   
    bool operator ==(const Buffer &aBuffer) const;
    
    inline bool operator !=(const Buffer &aBuffer) const
    {
        return ! (*this == aBuffer);
    }

    inline Buffer &operator = (const Buffer &aBuffer)    
    { 
        if(theRep == aBuffer.theRep) return *this;
        theRep->dismiss();
        theRep = aBuffer.theRep->copy();
        return *this; 
    }

    Buffer& operator+=(const Buffer &aBuffer);

    Buffer  operator+ (const Buffer &aBuffer)
    {
        Buffer newBuffer(*this);
        return newBuffer += aBuffer;
    }
};


#endif

