// Copyright (C) 2001 Fabio Lopiano <flopiano@bigfoot.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// Example program
//
// The message has the following structure:
//
// 1 byte     - number of objects "Person" following
// Object Person {
//    15 bytes name (null terminated)
//    Object Date {
//        1 byte  day
//        1 byte  month
//        2 byets year
//    }
//}
//
// a message with this structure is inserted in a buffer
// and extracted
//

#include <string>

#include <iostream>

#include "netobject.h"
#include "netvector.h"
#include "netarray.h"


class Date : public NetObject
{
    char  theDay;    
    char  theMonth;  
    short theYear;

protected:

    void netWrite( NetBuffer &aBuffer ) const
    {
        aBuffer << theDay << theMonth << theYear;
    }

    void netRead( NetBuffer &aBuffer ) 
    {
        aBuffer >> theDay >> theMonth >> theYear;
    }

public:
    Date() { netClear(); } // will set all members to zero

    // note that implementation details are hidden to the clients
    // and  all members are externally  treated as int

    Date( int d,int m, int y) :
	theDay(d), theMonth(m), theYear(y) {}

    int day() const { return theDay; }
    int month() const { return theMonth; }
    int year() const { return theYear; }
};


class Person : public NetObject
{
    enum { NameLength = 15 };

    NetArray<char,NameLength> theName;
    Date theBirthDate;
protected:

    void netWrite( NetBuffer &aBuffer ) const
    {
        aBuffer << theName << theBirthDate;
    }

    void netRead( NetBuffer &aBuffer ) 
    {
        aBuffer >> theName >> theBirthDate;
    }

public:
    Person() {} // NetClear() not needed: all members have default constructor

    Person (const std::string &aName, const Date &aDate ) :
	    theBirthDate(aDate)
    {
	    strncpy( theName, aName.c_str(), NameLength );
	    theName[ NameLength-1 ] = 0;
    }

    Date date() const { return theBirthDate; }
    std::string name() const { return std::string(theName); };
};

std::ostream & operator <<( std::ostream &os, const Date &d )
{
    return os << d.day() << '/' << d.month() << '/' << d.year();
}

std::ostream & operator <<( std::ostream &os, const Person &p )
{
    return os << p.name() << " " << p.date();
}


int main(int argc, const char **argv)
{
    typedef NetVector<Person,char> PersonVec;

    PersonVec myVec;

    myVec.push_back( Person("John Smith", Date( 18,7,1958)) );
    myVec.push_back( Person("Joe Hacker", Date( 14,5,1970)) );
    myVec.push_back( Person("Jack Lamer", Date( 8,11,1964)) );

    Buffer myBuf = myVec.toBuffer();

    // suppose to send / receive the buffer
    
    // the NetObjectT<> is not strictly needed here, 
    // because NetVector already has the constructor
    NetObjectT<PersonVec> myOtherVec (myBuf);      

    PersonVec::const_iterator it = myOtherVec.begin();
    for( ; it != myOtherVec.end(); ++it )
    {
        std::cout << (*it) << std::endl;
    }

    return 0;
}

